/*
 * Usage: anagram string [[min-len] wordfile]
 * Java Anagram program, Peter van der Linden Jan 7, 1996.
 * Feel free to pass this program around, as long as this
 * header stays intact.  
 */

import java.io.*;
import java.util.Vector;

interface UsefulConstants {
    public static final int MAXWORDS = 50000;
    public static final int MAXWORDLEN = 30;
    public static final int EOF = -1;

    // shorter alias for I/O streams
    public static final PrintStream o = System.out;
    public static final PrintStream e = System.err;
}

class Word {
    int mask;
    byte count[]= new byte[26];
    int total;
    String aword;

    Word(String s)  // construct an entry from a string
    {
        int ch;
        aword = s;
        mask = ~0;
        total = 0;
        s = s.toLowerCase();
        for (int i = 'a'; i <= 'z'; i++) count[i-'a'] = 0;

        for (int i = s.length()-1; i >= 0; i--) {
            ch = s.charAt(i) - 'a';
            if (ch >= 0 && ch < 26) {
                total++;
                count[ch]++;
                mask &= ~(1 << ch);
            }
        }
    }        
}

class WordList implements UsefulConstants {
    static Word[] Dictionary= new Word[MAXWORDS];
    static int totWords=0;

    static void ReadDict(String f)
    {
        FileInputStream fis;
        try fis = new FileInputStream(f);
        catch (FileNotFoundException fnfe) {
            e.println ("Cannot open the file of words '" + f + "'");
            throw new RuntimeException();
        }
        e.println("reading dictionary...");

        char buffer[] = new char[MAXWORDLEN];
        String s;
        int r =0;
        while (r!=EOF) {
            int i=0;
            try {
                // read a word in from the word file
                while ( (r=fis.read()) != EOF ) {
                // read past EOLN on a PC
                    if ( r == '\r' ) r=fis.read();
                    if ( r == '\n' ) break;
                    buffer[i++] = (char) r;
                }
            } catch (IOException ioe) {
                e.println ("Cannot read the file of words ");
                throw new RuntimeException();
            }
            s=new String(buffer,0,i);
            Dictionary[totWords] = new Word(s);
            totWords++;
        }

        e.println("main dictionary has " + totWords + " entries.");
    }


}

class anagram extends WordList implements UsefulConstants {

    static Word[] Candidate = new Word[MAXWORDS];
    static int totCandidates=0,
               MinimumLength = 3;

    public static void main(String[] argv) 
    {
        if ( argv.length < 1 || argv.length > 3) {
           e.println("Usage: anagram  string-to-anagram "
                     + "[min-len [word file]]");
           return;
        }
        if (argv.length >= 2) 
            MinimumLength = Integer.parseInt(argv[1]);

      // word filename is optional 3rd argument
        ReadDict( argv.length==3? argv[2] : "words.txt" );
        DoAnagrams(argv[0]);
    }


    static void DoAnagrams(String anag)
    {
        Word myAnagram = new Word(anag);

        myAnagram.mask = ~myAnagram.mask;

        getCandidates(myAnagram);

        int RootIndexEnd = sortCandidates(myAnagram);

        FindAnagram(myAnagram, new String[50] , 0, 0, RootIndexEnd);

        o.println("----" + anag + "----");
    }


    static void getCandidates(Word d) 
    {
        for (int i = totCandidates = 0; i < totWords; i++)
            if (   (  (Dictionary[i].mask | d.mask) == (int)~0)
                && (   Dictionary[i].total >= MinimumLength   )
                && (  (Dictionary[i].total + MinimumLength <= d.total)
                    || (Dictionary[i].total == d.total) )
                && ( fewerOfEachLetter(d.count, Dictionary[i].count) )  )

                Candidate[totCandidates++]=Dictionary[i];

        //e.println("Dictionary of words-that-are-substring-anagrams has "
        //           + totCandidates + " entries.");
        //PrintCandidate();
    }

    static boolean fewerOfEachLetter(byte anagCount[], byte entryCount[])
    {
        for (int i = 25; i >= 0; i--)
            if (entryCount[i] > anagCount[i]) return false;
        return true;
    }

    static void PrintCandidate()
    {
        for (int i = 0; i < totCandidates; i++)
            o.print( Candidate[i].aword + ", "
                     + ((i%4 == 3)?"\n":" " ) );
        o.println("");
    }

    static void FindAnagram(Word d, 
                            String WordArray[], 
                            int Level, int StartAt, int EndAt)
    {
        int i, j;
        boolean enoughCommonLetters;
        Word WordToPass = new Word("");

        for (i = StartAt; i < EndAt; i++) {
          if ( (d.mask | Candidate[i].mask) != 0) {
              enoughCommonLetters = true;
              for (j = 25; j >=0 && enoughCommonLetters; j--)
                  if (d.count[j] < Candidate[i].count[j]) 
                      enoughCommonLetters = false;

              if (enoughCommonLetters) {
                WordArray[Level] = Candidate[i].aword;
                WordToPass.mask = 0;
                WordToPass.total = 0;
                for (j = 25; j >= 0; j--) {
                    WordToPass.count[j] = (byte)
                     (    d.count[j] - 
                          Candidate[i].count[j] );
                    if ( WordToPass.count[j] != 0 ) {
                        WordToPass.total += 
                           (int)WordToPass.count[j];
                        WordToPass.mask |= 1 << j;
                    }
                }
                if (WordToPass.total == 0) {
                    /* Found a series of words! */ 
                    for (j = 0; j <= Level; j++) 
                        o.print(WordArray[j] + " ");
                    o.println();
                } else if (WordToPass.total < MinimumLength) {
                       ; /* Don't call again */
                } else {
                    FindAnagram(WordToPass, WordArray, Level+1, 
                         i, totCandidates);
                }
             }
         }
       }
    }

  
    static int SortMask;

    static int sortCandidates(Word d)
    {
        int [] MasterCount=new int[26];
        int LeastCommonIndex=0, LeastCommonCount;
        int i, j;

        for (j = 25; j >= 0; j--) MasterCount[j] = 0;

        for (i = totCandidates-1; i >= 0; i--)
            for (j = 25; j >= 0; j--)
                MasterCount[j] += Candidate[i].count[j];

        LeastCommonCount = MAXWORDS * 5;
        for (j = 25; j >= 0; j--)
            if (    MasterCount[j] != 0
                 && MasterCount[j] < LeastCommonCount 
                 && (d.mask & (1 << j) ) != 0  ) {
                LeastCommonCount = MasterCount[j];
                LeastCommonIndex = j;
            }

        SortMask = (1 << LeastCommonIndex);

        quickSort(0, totCandidates-1 );

        for (i = 0; i < totCandidates; i++)
            if ((SortMask & ~Candidate[i].mask) == 0) {
                break;
            }
        // e.println("least common letter is '"
        //            + (char)(LeastCommonIndex+'a') + "'" );
        // e.println("root breadth is " + i + " words");
        return i;
    }

    static void quickSort(int left, int right) 
    {
        // standard quicksort from any algorithm book
        int i, last;
        if (left >= right) return;
        swap(left, (left+right)/2);
        last = left;
        for (i=left+1; i<=right; i++)  /* partition */
            if (MultiFieldCompare( Candidate[i], 
                                   Candidate[left] ) == -1 )
                 swap( ++last, i);

        swap(last, left);
        quickSort(left, last-1);
        quickSort(last+1,right);
    }

    static int MultiFieldCompare(Word s, Word t)
    {
        if ( (s.mask & SortMask) != (t.mask & SortMask) )
            return ( (s.mask & SortMask)>(t.mask & SortMask)? 1:-1);

        if ( t.total != s.total )
            return (t.total - s.total);

        return (s.aword).compareTo(t.aword);
    }

    static void swap(int d1, int d2) {
        Word tmp = Candidate[d1];
        Candidate[d1] = Candidate[d2];
        Candidate[d2] = tmp;
    }
        
}

